use std::{
    collections::HashMap,
    ffi::OsString,
    fs::{create_dir, remove_dir_all, write, File, OpenOptions},
    io::Write,
    os::unix::ffi::OsStringExt,
    path::{Path, PathBuf},
    process::Command,
    time::SystemTime,
};

use radicle::git::Oid;

use clap::Parser;
use html_page::{Document, Element, Tag};
use log::{debug, error, info, trace};
use serde::Deserialize;
use tempfile::tempdir;
use time::{macros::format_description, OffsetDateTime};
use walkdir::WalkDir;

const STATS_TXT: &str = "stats.txt";
const SUCCESS: &str = "SUCCESS";
const FAILURE: &str = "FAILURE";

fn main() -> anyhow::Result<()> {
    pretty_env_logger::init_custom_env("WUMPUS_LOG");
    info!("I smell a Wumpus");

    let args = Args::parse();
    match &args.cmd {
        Cmd::Run(x) => x.run(&args)?,
        Cmd::Report(x) => x.run()?,
        Cmd::DummyLog(x) => x.run(),
    }

    info!("Wumpus hunter is done");
    Ok(())
}

fn dir_is_empty(dirname: &Path) -> bool {
    WalkDir::new(dirname)
        .min_depth(1)
        .into_iter()
        .next()
        .is_none()
}

#[derive(Debug, Parser)]
struct Args {
    #[clap(subcommand)]
    cmd: Cmd,
}

#[derive(Debug, Parser)]
enum Cmd {
    Run(Run),
    Report(ReportCmd),
    DummyLog(DummyLog),
}

#[derive(Debug, Parser)]
struct Run {
    #[clap(long)]
    repeats: Option<usize>,

    #[clap(long)]
    logs: PathBuf,

    #[clap(long, default_value = "600")]
    timeout: usize,

    #[clap(long)]
    check_tempdir: bool,

    spec: PathBuf,
}

impl Run {
    fn run(&self, args: &Args) -> anyhow::Result<()> {
        let spec = Spec::from_file(&self.spec)?;

        debug!("{args:#?}");
        debug!("{spec:#?}");

        assert!(self.logs.exists());
        let stats_txt = self.logs.join(STATS_TXT);

        let tmp = tempdir()?;
        let working_dir = tmp.path().join("srcdir");

        let report_html = self.logs.join("counts.html");
        let mut i = 0;
        loop {
            i += 1;
            if let Some(repeats) = &self.repeats {
                if i > *repeats {
                    break;
                }
            }
            info!("repetition {i}");

            let run_id = timestamp(&SystemTime::now());
            info!("run {run_id}");

            let mut run_log = RunLog::default();
            run_log.url(&spec.repository_url);

            if working_dir.exists() {
                spec.git_remote_update(&working_dir, &mut run_log)?;
                debug!("git pulled");
            } else {
                spec.git_clone(&working_dir, &mut run_log)?;
                debug!("git cloned");
            }
            spec.git_checkout(&working_dir, &spec.git_ref, &mut run_log)?;
            debug!("git checked out");

            let commit = spec.git_head(&working_dir, &mut run_log)?;

            let test_tmpdir = tmp.path().join("tmp");
            create_dir(&test_tmpdir)?;
            debug!("created temporary directory {}", test_tmpdir.display());

            let (mut log, mut success) =
                spec.run_test_suite(&working_dir, self.timeout, &test_tmpdir, &mut run_log)?;
            info!("ran test suite: {success}");

            if self.check_tempdir && !dir_is_empty(&test_tmpdir) {
                log.push_str("\n\n\n# Temporary files left behind\n");
                success = false;
                info!("test failed to clean up its temporary files: test failed");
            }

            remove_dir_all(&test_tmpdir)?;

            if success {
                record(&stats_txt, &commit, SUCCESS)?;
            } else {
                record(&stats_txt, &commit, FAILURE)?;
            }
            assert!(stats_txt.exists());

            let log_dir = self.logs.join(&format!("log-{}", commit));
            if !log_dir.exists() {
                create_dir(&log_dir)?;
            }
            let log_filename = log_dir.join(format!(
                "log-{run_id}.{i}.{}.html",
                if success { "success" } else { "fail" }
            ));
            write(&log_filename, run_log.as_html().to_string())?;

            let report = Report::new(&spec.description, &stats_txt, Path::new(&commit))?;
            write(
                &report_html,
                report.as_html(&spec, &working_dir).to_string(),
            )?;
        }

        Ok(())
    }
}

#[derive(Debug, Parser)]
struct ReportCmd {
    stats: PathBuf,
    spec: PathBuf,
    git: PathBuf,
}

impl ReportCmd {
    fn run(&self) -> anyhow::Result<()> {
        let spec = Spec::from_file(&self.spec)?;
        let report = Report::new(&spec.description, &self.stats, Path::new("/tmp"))?;
        print!("{}", report.as_html(&spec, &self.git));
        Ok(())
    }
}

#[derive(Debug, Deserialize)]
#[serde(deny_unknown_fields)]
struct Spec {
    description: String,
    repository_url: String,
    git_ref: String,
    command: String,
}

impl Spec {
    fn from_file(filename: &Path) -> anyhow::Result<Self> {
        let file = File::open(filename)?;
        let spec: Self = serde_yaml::from_reader(&file)?;
        Ok(spec)
    }

    fn git_clone(&self, working_dir: &Path, run_log: &mut RunLog) -> anyhow::Result<()> {
        RunCmd::new(".", run_log)
            .arg("git")
            .arg("clone")
            .arg(&self.repository_url)
            .path(working_dir)
            .run()?;
        Ok(())
    }

    fn git_remote_update(&self, working_dir: &Path, run_log: &mut RunLog) -> anyhow::Result<()> {
        RunCmd::new(working_dir, run_log)
            .arg("git")
            .arg("remote")
            .arg("update")
            .run()?;
        Ok(())
    }

    fn git_checkout(
        &self,
        working_dir: &Path,
        committish: &str,
        run_log: &mut RunLog,
    ) -> anyhow::Result<()> {
        RunCmd::new(working_dir, run_log)
            .arg("git")
            .arg("checkout")
            .arg(committish)
            .run()?;
        Ok(())
    }

    fn git_head(&self, working_dir: &Path, run_log: &mut RunLog) -> anyhow::Result<String> {
        let (stdout, _) = RunCmd::new(working_dir, run_log)
            .arg("git")
            .arg("rev-parse")
            .arg("HEAD")
            .run()?;
        Ok(stdout.trim().into())
    }

    fn git_commit_date(&self, working_dir: &Path, commit: &str, run_log: &mut RunLog) -> String {
        if let Ok((stdout, _)) = RunCmd::new(working_dir, run_log)
            .arg("git")
            .arg("show")
            .arg("--pretty=fuller")
            .arg("--date=iso")
            .arg(commit)
            .run()
        {
            const PREFIX: &str = "CommitDate: ";
            let x: Vec<String> = stdout
                .lines()
                .filter_map(|line| line.strip_prefix(PREFIX).map(|s| s.to_string()))
                .collect();
            if x.len() == 1 {
                return x[0].clone();
            }
        }
        "(unknown date)".into()
    }

    fn run_test_suite(
        &self,
        working_dir: &Path,
        timeout: usize,
        tmpdir: &Path,
        run_log: &mut RunLog,
    ) -> anyhow::Result<(String, bool)> {
        RunCmd::new(working_dir, run_log)
            .tmpdir(tmpdir)
            .arg("timeout")
            .arg(&format!("{}s", timeout))
            .arg("bash")
            .arg("-c")
            .arg(&self.command)
            .run()
    }
}

#[derive(Debug)]
struct RunCmd<'a> {
    argv: Vec<OsString>,
    cwd: PathBuf,
    tmpdir: Option<PathBuf>,
    run_log: &'a mut RunLog,
}

impl<'a> RunCmd<'a> {
    fn new<P: AsRef<Path>>(cwd: P, run_log: &'a mut RunLog) -> Self {
        let cwd = cwd.as_ref();
        if !cwd.exists() {
            error!("ERROR: directory {} does not exist", cwd.display());
        }
        assert!(cwd.exists());
        Self {
            argv: vec![],
            cwd: cwd.into(),
            tmpdir: None,
            run_log,
        }
    }

    fn tmpdir(mut self, dirname: &Path) -> Self {
        self.tmpdir = Some(dirname.into());
        self
    }

    fn arg(mut self, arg: &str) -> Self {
        self.argv.push(OsString::from_vec(arg.as_bytes().to_vec()));
        self
    }

    fn path(mut self, arg: &Path) -> Self {
        self.argv.push(arg.as_os_str().into());
        self
    }

    fn run(self) -> anyhow::Result<(String, bool)> {
        trace!("runcmd: {self:#?}");
        let argv0 = self.argv.first().unwrap();
        let args = &self.argv[1..];
        let tmpdir = self.tmpdir.unwrap_or(PathBuf::from("/tmp"));
        let output = Command::new(argv0)
            .args(args)
            .current_dir(&self.cwd)
            .env("TMPDIR", &tmpdir)
            .output()?;
        let stdout = String::from_utf8_lossy(&output.stdout).to_string();
        trace!("runcmd: stdout:\n{stdout}");
        let stderr = String::from_utf8_lossy(&output.stderr).to_string();
        trace!("runcmd: stderr:\n{stderr}");
        trace!("runcmd: success? {}", output.status.success());

        let argv: Vec<&str> = self.argv.iter().map(|os| os.to_str().unwrap()).collect();

        self.run_log
            .runcmd(&argv, output.status.code().unwrap(), &stdout, &stderr);
        Ok((stdout, output.status.success()))
    }
}

fn record(filename: &Path, commit: &str, result: &str) -> anyhow::Result<()> {
    trace!("writing {result} on {commit} to {}", filename.display());
    let mut file = OpenOptions::new()
        .create(true)
        .append(true)
        .open(filename)?;
    file.write_all(format!("{commit} {result}\n").as_bytes())?;
    Ok(())
}

struct Report {
    description: String,
    commits: Vec<(String, bool)>,
    log_dir: PathBuf,
}

impl Report {
    const CSS: &'static str = r#"
h1 { font-size: 200%; }
h2 { font-size: 125%; }
h3 { font-size: 100%; }

table {
    width: 100%;
    text-align: left;
    tr:nth-child(even) {background-color: #f2f2f2;}
}

.commit {
    font-family: monospace;
}

.numeric {
    text-align: right;
}
"#;

    fn new(description: &str, stats_txt: &Path, log_dir: &Path) -> anyhow::Result<Self> {
        let text = std::fs::read(stats_txt)?;
        let text = String::from_utf8_lossy(&text);
        let mut commits = vec![];
        for line in text.lines() {
            let mut words = line.split_whitespace();
            let commit = words.next().unwrap();
            let result = words.next().unwrap();
            commits.push((commit.to_string(), result == "SUCCESS"));
        }
        trace!("loaded {} commits", commits.len());
        Ok(Self {
            description: description.into(),
            commits,
            log_dir: log_dir.into(),
        })
    }

    fn count(&self) -> HashMap<String, (usize, usize)> {
        let mut counts = HashMap::new();
        for (commit, success) in self.commits.iter() {
            let (mut s, mut f) = counts.get(commit).copied().unwrap_or((0, 0));
            if *success {
                s += 1;
            } else {
                f += 1;
            }
            counts.insert(commit.into(), (s, f));
        }
        counts
    }

    fn counts_with_date(
        spec: &Spec,
        working_dir: &Path,
        counts: HashMap<String, (usize, usize)>,
    ) -> HashMap<String, (String, usize, usize)> {
        let mut map = HashMap::new();
        let mut run_log = RunLog::default();
        for (commit, (successes, failures)) in counts.iter() {
            let date = spec.git_commit_date(working_dir, commit, &mut run_log);
            map.insert(commit.to_string(), (date, *successes, *failures));
        }
        map
    }

    fn as_html(&self, spec: &Spec, working_dir: &Path) -> Document {
        let counts = self.count();
        let counts = Self::counts_with_date(spec, working_dir, counts);

        let now = timestamp(&SystemTime::now());

        let mut table = Element::new(Tag::Table).with_child(
            Element::new(Tag::Tr)
                .with_child(Element::new(Tag::Th).with_text("date"))
                .with_child(Element::new(Tag::Th).with_text("commit"))
                .with_child(
                    Element::new(Tag::Th)
                        .with_class("numeric")
                        .with_text("successes"),
                )
                .with_child(
                    Element::new(Tag::Th)
                        .with_class("numeric")
                        .with_text("failures"),
                ),
        );

        let mut counts: Vec<_> = counts.iter().collect();
        counts.sort_by_cached_key(|(_, (date, _, _))| date);
        counts.reverse();

        for (commit, (date, successes, failures)) in counts.iter() {
            let successes = format!("{successes}");
            let failures = format!("{failures}");

            table.push_child(
                Element::new(Tag::Tr)
                    .with_child(Element::new(Tag::Td).with_class("date").with_text(date))
                    .with_child(
                        Element::new(Tag::Td).with_child(
                            Element::new(Tag::A)
                                .with_attribute("href", &format!("{}/", self.log_dir.display()))
                                .with_class("commit")
                                .with_text(commit),
                        ),
                    )
                    .with_child(
                        Element::new(Tag::Td)
                            .with_class("numeric")
                            .with_text(&successes),
                    )
                    .with_child(
                        Element::new(Tag::Td)
                            .with_class("numeric")
                            .with_text(&failures),
                    ),
            );
        }

        let total = self.commits.len();
        Document::default()
            .with_head_element(Element::new(Tag::Title).with_text("Wumpus hunter"))
            .with_head_element(Element::new(Tag::Style).with_text(Self::CSS))
            .with_body_element(Element::new(Tag::H1).with_text("Wumpus hunter"))
            .with_body_element(Element::new(Tag::P).with_text(&self.description))
            .with_body_element(
                Element::new(Tag::P)
                    .with_text(&format!("Total of {total} test runs. Last updated "))
                    .with_text(&now),
            )
            .with_body_element(table)
    }
}

fn timestamp(when: &SystemTime) -> String {
    let fmt = format_description!("[year]-[month]-[day] [hour]:[minute]:[second]Z");
    OffsetDateTime::from(*when).format(fmt).ok().unwrap()
}

#[derive(Debug, Parser)]
struct DummyLog {}

impl DummyLog {
    fn run(&self) {
        let mut run_log = RunLog::default();
        run_log.url("https://liw.fi");
        run_log.git_ref("master");
        run_log.git_commit(Oid::try_from("06f52d8a538e79b09bd8252a1346542ed179cd60").unwrap());
        run_log.runcmd(
            &["git", "clean", "-fdx"],
            0,
            "This is stdout",
            "This is stderr",
        );
        println!("{}", run_log.as_html());
    }
}

#[derive(Debug)]
struct RunCmdLog {
    argv: Vec<String>,
    exit: i32,
    stdout: String,
    stderr: String,
}

impl RunCmdLog {
    fn heading(&self) -> String {
        let mut h = String::new();
        for arg in self.argv.iter() {
            h.push_str(arg);
            h.push(' ');
        }
        h
    }
}

#[derive(Default, Debug)]
struct RunLog {
    url: Option<String>,
    git_ref: Option<String>,
    git_commit: Option<Oid>,
    runcmds: Vec<RunCmdLog>,
}

impl RunLog {
    const TITLE: &'static str = "Wumpus run log";

    fn url(&mut self, url: &str) {
        self.url = Some(url.into());
    }

    fn git_ref(&mut self, git_ref: &str) {
        self.git_ref = Some(git_ref.into());
    }

    fn git_commit(&mut self, commit: Oid) {
        self.git_commit = Some(commit);
    }

    fn runcmd(&mut self, argv: &[&str], exit: i32, stdout: &str, stderr: &str) {
        self.runcmds.push(RunCmdLog {
            argv: argv.iter().map(|s| s.to_string()).collect(),
            exit,
            stdout: stdout.into(),
            stderr: stderr.into(),
        });
    }

    fn as_html(&self) -> Document {
        let url = if let Some(url) = &self.url {
            url.as_ref()
        } else {
            "unknown URL"
        };

        let git_ref = if let Some(git_ref) = &self.git_ref {
            Element::new(Tag::Code).with_text(git_ref)
        } else {
            Element::new(Tag::Span).with_text("unknown ref")
        };

        let commit = if let Some(commit) = &self.git_commit {
            Element::new(Tag::Code).with_text(&commit.to_string())
        } else {
            Element::new(Tag::Span).with_text("unknown commit")
        };

        let mut doc = Document::default()
            .with_head_element(Element::new(Tag::Title).with_text(Self::TITLE))
            .with_body_element(Element::new(Tag::H1).with_text(Self::TITLE))
            .with_body_element(Element::new(Tag::H2).with_text("Metadata"))
            .with_body_element(
                Element::new(Tag::Ul)
                    .with_child(
                        Element::new(Tag::Li).with_text("Repository: ").with_child(
                            Element::new(Tag::A)
                                .with_attribute("href", url)
                                .with_child(Element::new(Tag::Code).with_text(url)),
                        ),
                    )
                    .with_child(
                        Element::new(Tag::Li)
                            .with_text("Git ref: ")
                            .with_child(git_ref),
                    )
                    .with_child(
                        Element::new(Tag::Li)
                            .with_text("Commit: ")
                            .with_child(commit),
                    ),
            );

        let mut toc = Element::new(Tag::Ul);
        let mut cmds = vec![];

        for (i, run) in self.runcmds.iter().enumerate() {
            let id = format!("run{i}");
            let argv_id = format!("run{i}-argv");
            let exit_id = format!("run{i}-exit");
            let stdout_id = format!("run{i}-stdout");
            let stderr_id = format!("run{i}-stderr");

            cmds.push(
                Element::new(Tag::H2)
                    .with_attribute("id", &id)
                    .with_text("Run: ")
                    .with_text(&run.heading()),
            );
            toc.push_child(
                Element::new(Tag::Li).with_child(
                    Element::new(Tag::A)
                        .with_attribute("href", &format!("#{id}"))
                        .with_text(&run.heading()),
                ),
            );

            let mut sub_toc = Element::new(Tag::Ul);

            cmds.push(
                Element::new(Tag::H3)
                    .with_text("Argv")
                    .with_attribute("id", &argv_id),
            );
            sub_toc.push_child(
                Element::new(Tag::Li).with_child(
                    Element::new(Tag::A)
                        .with_attribute("href", &format!("#{argv_id}"))
                        .with_text("Argv"),
                ),
            );
            let mut list = Element::new(Tag::Ol);
            for arg in run.argv.iter() {
                list.push_child(
                    Element::new(Tag::Li).with_child(Element::new(Tag::Code).with_text(arg)),
                );
            }
            cmds.push(list);

            cmds.push(
                Element::new(Tag::H3)
                    .with_text("Exit code")
                    .with_attribute("id", &exit_id),
            );
            sub_toc.push_child(
                Element::new(Tag::Li).with_child(
                    Element::new(Tag::A)
                        .with_attribute("href", &format!("#{exit_id}"))
                        .with_text("Exit code"),
                ),
            );
            cmds.push(Element::new(Tag::P).with_text(&format!("Exit code {}", run.exit)));

            if !run.stdout.is_empty() {
                cmds.push(
                    Element::new(Tag::H3)
                        .with_text("Stdout")
                        .with_attribute("id", &stdout_id),
                );
                sub_toc.push_child(
                    Element::new(Tag::Li).with_child(
                        Element::new(Tag::A)
                            .with_attribute("href", &format!("#{stdout_id}"))
                            .with_text("Stdout"),
                    ),
                );
                cmds.push(
                    Element::new(Tag::Blockquote)
                        .with_child(Element::new(Tag::Pre).with_text(&run.stdout)),
                );
            }

            if !run.stderr.is_empty() {
                cmds.push(
                    Element::new(Tag::H3)
                        .with_text("Stderr")
                        .with_attribute("id", &stderr_id),
                );
                sub_toc.push_child(
                    Element::new(Tag::Li).with_child(
                        Element::new(Tag::A)
                            .with_attribute("href", &format!("#{stderr_id}"))
                            .with_text("Stderr"),
                    ),
                );
                cmds.push(
                    Element::new(Tag::Blockquote)
                        .with_child(Element::new(Tag::Pre).with_text(&run.stderr)),
                );
            }

            toc.push_child(sub_toc);
        }

        doc.push_to_body(Element::new(Tag::H2).with_text("Contents"));
        doc.push_to_body(toc);
        for e in cmds {
            doc.push_to_body(e);
        }

        doc
    }
}
